///
/// \file packager.cc
/// \author K. Isom <kyle@imap.cc>
/// \date 2023-10-18
/// \brief Packaging tools for shimmering-clarity.
///

#include <filesystem>
#include <iostream>
#include <string>

#include <curlpp/Easy.hpp>
#include <curlpp/Exception.hpp>
#include <curlpp/Options.hpp>
#include <curlpp/cURLpp.hpp>

#include <fstream>
#include <scsl/Commander.h>
#include <scsl/Flags.h>
#include <scsl/SimpleConfig.h>


const auto defaultOwner        = std::string("sc");
const auto defaultDistribution = std::string("ubuntu");
const auto defaultComponent    = std::string("main");

void
usage(std::ostream &outs, int exc)
{
	outs << "sc3dev-packager: manage packages on gitea\n";
	outs << "usage:\n";
	outs << "\tsc3dev-packager [--component component] [--owner owner] \n";
	outs << "\t\t[--distribution distribution] debian command packages ...\n";
	outs << "\tsc3dev-packager [--owner owner] generic command packages ...\n";
	outs << "\nCommands:\n";
	outs << "\t- delete: remove a package\n";
	outs << "\t- put: upload a package\n";

	exit(exc);
}


std::string
getDefaultConfigFile()
{
	auto userEnv = getenv("USER");
	auto user    = std::string(userEnv);

	if (user == "root") {
		return std::filesystem::path("/etc/sc3dev/packager.conf");
	}

	if (user.empty()) {
		return "";
	}

	userEnv       = getenv("HOME");
	auto userHome = std::filesystem::path(userEnv);

	userHome /= std::filesystem::path(".config");
	userHome /= std::filesystem::path("sc3dev-packager.conf");
	return userHome;
}


static std::string
debianEndpoint(std::string endpoint)
{
	if (!endpoint.empty() && endpoint.at(0) != '/') {
		endpoint.insert(endpoint.cbegin(), '/');
	}

	auto urlBase = scsl::SimpleConfig::GetGlobal("server");
	return urlBase + "/api/packages/" +
	       scsl::SimpleConfig::GetGlobal("owner", defaultOwner) +
	       "/debian/pool/" +
	       scsl::SimpleConfig::GetGlobal("distribution", defaultDistribution) +
	       "/" +
	       scsl::SimpleConfig::GetGlobal("component", defaultComponent) +
	       endpoint;
}

static std::string
getUserPass()
{
	return scsl::SimpleConfig::GetGlobal("http_user") + ":" +
	       scsl::SimpleConfig::GetGlobal("http_password");
}

static bool
putDebian(std::vector<std::string> args)
{
	auto ok = true;
	std::cout << "[+] put Debian package\n";
	for (auto &arg: args) {
		std::cout << "\t[*] " << arg << "\n";
	}

	auto url      = debianEndpoint("upload");
	auto userPass = getUserPass();
	for (auto &arg: args) {
		try {
			std::list<std::string> header;
			header.emplace_back("Content-Type: application/octet-stream");

			curlpp::Cleanup cleaner;
			curlpp::Easy request;

			if (!std::filesystem::exists(arg)) {
				std::cerr << "[!] " << arg << " does not exist!\n";
				ok &= false;
				continue;
			}
			auto fileSize = static_cast<long>(std::filesystem::file_size(arg));

			std::ifstream debFile(arg);
			if (!debFile.good()) {
				std::cerr << "[!] " << arg << " does not exist!\n";
				ok &= false;
				continue;
			}

			request.setOpt(new curlpp::options::Url(url));
			request.setOpt(new curlpp::options::Verbose(true));
			request.setOpt(new curlpp::options::UserPwd(userPass));
			request.setOpt(new curlpp::Options::ReadStream(&debFile));
			request.setOpt(new curlpp::Options::InfileSize(fileSize));
			request.setOpt(new curlpp::Options::Upload(true));
			request.perform();
		} catch (curlpp::LogicError &e) {
			std::cout << e.what() << "\n";
			ok &= false;
		} catch (curlpp::RuntimeError &e) {
			std::cerr << e.what() << "\n";
			ok &= false;
		}

		std::cout << "[+] " << std::filesystem::path(arg).filename() << ": OK\n";
	}

	return false;
}


static bool
debianPackager(std::vector<std::string> args)
{
	scsl::Commander commander;
	commander.Register(scsl::Subcommand("put", 1, putDebian));

	std::cout << "[+] Debian packager invoked\n";
	for (auto &arg: args) {
		std::cout << "\t[*] " << arg << "\n";
	}

	auto command = args[0];
	args.erase(args.cbegin());
	return commander.Run(command, args) == scsl::Subcommand::Status::OK;
}


int
main(int argc, char *argv[])
{
	int retc;
	scsl::Flags *flags;
	std::string configFile = getDefaultConfigFile();
	std::string owner;
	std::string distribution;
	std::string component;
	std::string version;

	flags = new scsl::Flags("sc3dev-packager", "package tooling for shimmering clarity");
	flags->Register("--config", configFile, "path to a configuration file");
	flags->Register("--component", "main", "Debian package component");
	flags->Register("--distribution", "ubuntu", "Debian package distribution");
	flags->Register("--owner", "sc", "package repository owner");
	flags->Register("--version", scsl::FlagType::String, "version to operate on (only matters for delete)");

	auto status = flags->Parse(argc, argv);
	if (status != scsl::Flags::ParseStatus::OK) {
		std::cerr << "failed to parse flags: "
			  << scsl::Flags::ParseStatusToString(status)
			  << "\n";
		exit(1);
	}

	scsl::Commander commander;
	commander.Register(scsl::Subcommand("debian", 1, debianPackager));

	flags->GetString("--config", configFile);
	if (std::filesystem::exists(std::filesystem::path(configFile))) {
		if (scsl::SimpleConfig::LoadGlobal(configFile) != 0) {
			std::cerr << "[!] failed to read config file " << configFile << ".\n";
			exit(1);
		}
	}

	if (flags->GetString("--owner", owner)) {
		scsl::SimpleConfig::PutGlobal("owner", owner);
	}

	if (flags->GetString("--distribution", distribution)) {
		scsl::SimpleConfig::PutGlobal("distribution", distribution);
	}

	if (flags->GetString("--component", component)) {
		scsl::SimpleConfig::PutGlobal("component", component);
	}

	if (flags->GetString("--version", version)) {
		scsl::SimpleConfig::PutGlobal("version", version);
	}

	if (flags->NumArgs() < 1) {
		std::cerr << "[!] not enough arguments provided.\n";
		exit(1);
	}

	auto command = flags->Arg(0);
	auto args    = flags->Args();
	args.erase(args.begin());

	auto result = commander.Run(command, args);
	delete flags;

	switch (result) {
	case scsl::Subcommand::Status::OK:
		std::cout << "[+] OK\n";
		retc = 0;
		break;
	case scsl::Subcommand::Status::NotEnoughArgs:
		usage(std::cerr, 1);
		break;
	case scsl::Subcommand::Status::Failed:
		std::cerr << "[!] '" << command << "' failed\n";
		break;
	case scsl::Subcommand::Status::CommandNotRegistered:
		std::cerr << "[!] '" << command << "' not registered.\n";
		usage(std::cerr, 1);
		break;
	default:
		abort();
	}

	return retc;
}
